"""Modules for working with exotic options.

Exotic options can refer to auto-callables, barrier options, etc."""

from exotx.instruments.autocallable import Autocallable

__all__ = [
    'Autocallable',
]